!-------------------------------------------------------------------------------
! Description:
!
!   Read RTTOV coefficients and allocate data structures.
!
! Copyright:
!
!   This software was developed within the context of the EUMETSAT Satellite
!   Application Facility on Numerical Weather Prediction (NWP SAF), under the
!   Cooperation Agreement dated 29 June 2011, between EUMETSAT and the
!   Met Office, UK, by one or more partners within the NWP SAF. The partners
!   in the NWP SAF are the Met Office, ECMWF, KNMI and MeteoFrance.
!
!   Copyright 2017, EUMETSAT, All Rights Reserved.
!
!-------------------------------------------------------------------------------

subroutine radsim_init_rttov_data( &
  alloc,          & ! in
  nchans,         & ! in
  nlevels,        & ! in
  opts,           & ! in
  opts_scatt,     & ! in
  coeffs,         & ! in
  scat_coeffs,    & ! in
  chanprof,       & ! out
  radiance,       & ! out
  radiance_k,     & ! out
  transmission,   & ! out
  transmission_k, & ! out
  emissivity,     & ! out
  emissivity_k,   & ! out
  profiles,       & ! out
  profiles_k,     & ! out
  cld_profiles,   & ! out
  frequencies     ) ! out

use parkind1, only : &
  jpim, jplm

use radsim_mod_cfg

use radsim_mod_constants, only : &
  status_error, &
  output_debug

use radsim_mod_process, only : &
  nprofs_per_call

use rttov_const, only : &
  errorstatus_success

use rttov_types, only : &
  rttov_options,     &
  rttov_options_scatt, &
  rttov_coefs,       &
  rttov_scatt_coef, &
  rttov_profile,      &
  rttov_profile_cloud, &
  rttov_transmission, &
  rttov_radiance,     &
  rttov_chanprof, &
  rttov_emissivity

implicit none

include 'radsim_error_report.interface'
#include "rttov_alloc_prof.interface"
#include "rttov_alloc_rad.interface"
#include "rttov_alloc_scatt_prof.interface"
#include "rttov_alloc_transmission.interface"
#include "rttov_dealloc_coefs.interface"
#include "rttov_errorreport.interface"
#include "rttov_scatt_setupindex.interface"
!#include "rttov_user_options_checkinput.interface"

integer(kind=jpim),                  intent(in)    :: alloc ! 1=allocate, 0=deallocate
integer(kind=jpim),                  intent(in)    :: nchans
integer(kind=jpim),                  intent(in)    :: nlevels
type(rttov_options),                 intent(in)    :: opts
type(rttov_options_scatt),           intent(in)    :: opts_scatt
type(rttov_coefs),                   intent(in)    :: coeffs(:)
type(rttov_scatt_coef),              intent(in)    :: scat_coeffs
type(rttov_chanprof), allocatable,   intent(inout)   :: chanprof(:)
type(rttov_radiance),                intent(inout)   :: radiance
type(rttov_radiance),                intent(inout)   :: radiance_k
type(rttov_transmission),            intent(inout)   :: transmission
type(rttov_transmission),            intent(inout)   :: transmission_k
type(rttov_emissivity), allocatable, intent(inout)   :: emissivity(:)
type(rttov_emissivity), allocatable, intent(inout)   :: emissivity_k(:)
type(rttov_profile), allocatable,    intent(inout)   :: profiles(:)
type(rttov_profile), allocatable,    intent(inout)   :: profiles_k(:)
type(rttov_profile_cloud), allocatable, intent(inout) :: cld_profiles(:)
integer(kind=jpim), allocatable,     intent(inout)   :: frequencies(:)

integer, parameter :: nrttovid = 1        ! Number of sensor coeff files to read

character(len=400) :: file_name_clear
character(len=400) :: file_name_cloud
character(len=80) :: message
integer :: i, j, chan
integer(kind=jpim) :: nprofchans
integer(kind=jpim) :: errorstatus
integer(kind=jpim) :: file_unit
integer(kind=jpim) :: atlas_type
logical(kind=jplm), allocatable :: usechans(:,:)

!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! 1. Set up indices
!-------------------------------------------------------------------------------

nprofchans = nchans * nprofs_per_call

if ( alloc == 1 ) then

  if ( output_mode >= output_debug ) then
    print '(a)', ' Setting up channel arrays'
  end if

  allocate(chanprof(nprofchans))

  if ( .not. run_scatt ) then

    ! Build the list of chanprof indices (doesn't take into account that rttov may
    ! define channels differently).

    do j = 1, nprofs_per_call
      do i = 1, nchans
        chanprof((j-1)*nchans+i) % prof = j
        chanprof((j-1)*nchans+i) % chan = channels(i)
      end do
    end do

  else

    allocate(frequencies(nprofchans))
    frequencies = 0

    allocate(usechans(nprofs_per_call, coeffs(1) % coef % fmv_chn))
    usechans = .false.
    do i = 1, nchans
      chan = channels(i)
      usechans(:,chan) = .true.
    end do

    call rttov_scatt_setupindex ( &
      nprofs_per_call,            & ! in
      coeffs(1) % coef % fmv_chn, & ! in
      coeffs(1),                  & ! in
      nprofchans,                 & ! in
      chanprof,                   & ! out
      frequencies,                & ! out
      lchannel_subset=usechans    ) ! in, optional

    deallocate(usechans)

  end if

else

  deallocate(chanprof)
  if ( allocated(frequencies) ) deallocate(frequencies)

end if

!-------------------------------------------------------------------------------
! 2. Allocate other RTTOV structures
!-------------------------------------------------------------------------------

!----------------
! 2.1 Direct code
!----------------

! Profile structure

if ( alloc == 1 ) then
  if ( output_mode >= output_debug ) then
    print '(a)', ' Allocating RTTOV structures'
  end if
  allocate(profiles(nprofs_per_call))
  profiles(:) % nlevels = nlevels
  profiles(:) % nlayers = nlevels-1
end if

call rttov_alloc_prof( &
  errorstatus,       &
  nprofs_per_call,   &
  profiles,          &
  nlevels,           &
  opts,              &
  alloc,             &
  coefs = coeffs(1), &
  init = .true._jplm ) ! zero contents?

if ( errorstatus /= errorstatus_success ) then
  write(message, '(a,i0)') 'Error in rttov_alloc_prof = ', errorstatus
  call radsim_error_report(message, status_error)
end if

if ( alloc == 0 ) then
  deallocate(profiles)
end if

! Radiance structure

call rttov_alloc_rad( &
  errorstatus, &
  nprofchans,  &
  radiance,    &
  nlevels,     &
  alloc        ) ! 1=allocate, 0=deallocate

if ( errorstatus /= errorstatus_success ) then
  write(message, '(a,i0)') 'Error in rttov_alloc_rad = ', errorstatus
  call radsim_error_report(message, status_error)
end if

! Transmittance structure

call rttov_alloc_transmission( &
  errorstatus,       &
  transmission,      &
  nlevels,           &
  nprofchans,        &
  alloc,             & ! 1=allocate, 0=deallocate
  init = .true._jplm )

if ( errorstatus /= errorstatus_success ) then
  write(message, '(a,i0)') 'Error in rttov_alloc_transmission = ', errorstatus
  call radsim_error_report(message, status_error)
end if

! Emissivity structure

if ( alloc == 1 ) then
  allocate(emissivity(nprofchans))
  emissivity % emis_in = 0.0
  emissivity % emis_out = 0.0
else
  deallocate(emissivity)
end if

!-----------
! 4.2 K code
!-----------

if ( run_k ) then

  ! Profile structure

  if ( alloc == 1 ) then
    if ( output_mode >= output_debug ) then
      print '(a)', ' Allocating RTTOV K structures'
    end if
    allocate(profiles_k(nprofchans))
    profiles_k(:) % nlevels = nlevels
    profiles_k(:) % nlayers = nlevels-1
  end if

  call rttov_alloc_prof( &
    errorstatus,         &
    nprofchans,          &
    profiles_k,          &
    nlevels,             &
    opts,                &
    alloc,               & ! 1=allocate, 0=deallocate
    coefs = coeffs(1),   &
    init = .true._jplm   ) ! zero contents?

  if ( errorstatus /= errorstatus_success ) then
    write(message, '(a,i0)') 'Error in rttov_alloc_prof = ', errorstatus
    call radsim_error_report(message, status_error)
  end if

  if ( alloc == 0 ) then
    deallocate(profiles_k)
  end if

  ! Radiance structure

  call rttov_alloc_rad( &
    errorstatus,       &
    nprofchans,        &
    radiance_k,        &
    nlevels,           &
    alloc,             & ! 1=allocate, 0=deallocate
    init = .true._jplm )

  if ( alloc == 1 ) radiance_k % bt = 1.0

  if ( errorstatus /= errorstatus_success ) then
    write(message, '(a,i0)') 'Error in rttov_alloc_rad = ', errorstatus
    call radsim_error_report(message, status_error)
  end if

  ! Transmittance structure

  call rttov_alloc_transmission( &
    errorstatus,       &
    transmission_k,    &
    nlevels,           &
    nprofchans,        &
    alloc,             & ! 1=allocate, 0=deallocate
    init = .true._jplm )

  if ( errorstatus /= errorstatus_success ) then
    write(message, '(a,i0)') 'Error in rttov_alloc_transmission = ', errorstatus
    call radsim_error_report(message, status_error)
  end if

  if ( alloc == 1 ) then
    allocate(emissivity_k(nprofchans))
    emissivity_k % emis_in = 0.0
    emissivity_k % emis_out = 0.0
  else
    deallocate(emissivity_k)
  end if

end if

!---------------------------
! 4.3 Scattering direct code
!---------------------------

if ( run_scatt ) then

  if ( alloc == 1 ) then
    allocate(cld_profiles(nprofs_per_call))
  end if

  call rttov_alloc_scatt_prof( &
    errorstatus, &
    nprofs_per_call, &
    cld_profiles,    &
    nlevels,         &
    transfer(use_totalice, .true._jplm), & ! use totalice?
    alloc,           & ! 1=allocate, 0=deallocate
    init=.true._jplm ) ! zero contents?

  if ( alloc == 0 ) then
    deallocate(cld_profiles)
  end if

end if

end subroutine radsim_init_rttov_data
