!-------------------------------------------------------------------------------
! Description:
!
!   Read a GRIB file into a model structure.
!
! Copyright:
!
!   This software was developed within the context of the EUMETSAT Satellite
!   Application Facility on Numerical Weather Prediction (NWP SAF), under the
!   Cooperation Agreement dated 29 June 2011, between EUMETSAT and the
!   Met Office, UK, by one or more partners within the NWP SAF. The partners
!   in the NWP SAF are the Met Office, ECMWF, KNMI and MeteoFrance.
!
!   Copyright 2017, EUMETSAT, All Rights Reserved.
!
!-------------------------------------------------------------------------------

subroutine radsim_read_grib(file_name, model)

use grib_api

use radsim_mod_cfg, only : &
  output_mode, &
  run_scatt, &
  mmr_snowrain, &
  clw_data, &
  ozone_data, &
  ir_addclouds

use radsim_mod_constants, only : &
  real64, &
  real32, &
  int32, &
  gravity, &
  status_warning, &
  status_error, &
  output_verbose, &
  output_default

use radsim_mod_process

use radsim_mod_types, only : &
  model_type

implicit none

include 'radsim_calc_ecmwf_plevels.interface'
include 'radsim_error_report.interface'
include 'radsim_grib_paramid_name.interface'
include 'radsim_grid_calc.interface'

! Subroutine args
character(len=400), intent(in)  :: file_name
type(model_type),   intent(out) :: model

! Local variables
integer :: i, j
integer(int32) :: file_id, message_id, status
integer :: nmessage
integer :: field_list(nfields_rttov, 3)
logical :: got_field(nfields_rttov)
integer :: nprofs, nlevels
integer :: field_index
integer :: paramId
integer :: numberOfDataPoints
integer :: NV
real(real64), allocatable :: pv(:)
integer :: gridType
integer :: Ni
integer :: Nj
real :: latitudeOfFirstGridPointInDegrees
real :: longitudeOfFirstGridPointInDegrees
real :: latitudeOfLastGridPointInDegrees
real :: longitudeOfLastGridPointInDegrees
real :: jDirectionIncrementInDegrees
real :: iDirectionIncrementInDegrees
integer :: jPointsAreConsecutive
integer :: iScansNegatively
integer :: jScansPositively
character(len=10) :: centre = ''
character(len=10) :: dataDate
character(len=6) :: dataTime
character(len=80) :: message
integer :: level
real(real64), pointer :: values(:)
real(real64) :: missingvalue
character(len=20) :: param_name

!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! 1. Define required field list
!-------------------------------------------------------------------------------

! Each field has a unique ID which is the same in GRIB-1 and 2. These are
! stored in the paramID variable.

! Some fields may be stored incorrectly, or they may use different level codes
! to that of the UM, hence we allow for primary and secondary requirements
! (secondary used if primary not present). Also, some fields may be derived
! from secondary fields under certain circumstances.

field_list = -1

! See Annex in RTTOV Users Guide for required fields

!---------------
! Surface fields
!---------------

field_list(field_rttov_tskin,    1) = 235  ! skin temperature
field_list(field_rttov_t2,       1) = 167  ! temperature at 2m
field_list(field_rttov_q2,       1) = 168  ! dewpoint temperature at 2m
field_list(field_rttov_pstar,    1) = 134  ! surface pressure
field_list(field_rttov_pstar,    2) = 152  ! log of surface pressure
field_list(field_rttov_u10,      1) = 165  ! 10m wind u component
field_list(field_rttov_v10,      1) = 166  ! 10m wind v component

! Surface type usually needs to be derived from 2 fields - a land-sea mask and
! the seaice fraction. The priority is disabled in this case.

field_list(field_rttov_surftype, 1) = 172  ! LSM
field_list(field_rttov_surftype, 2) =  31  ! seaice cover
field_list(field_rttov_surftype, 3) = 129  ! geopotential

!---------
! Profiles
!---------

field_list(field_rttov_p,   1) =  54  ! pressure levels
field_list(field_rttov_t,   1) = 130  ! T
field_list(field_rttov_t,   2) =   3  ! theta
field_list(field_rttov_q,   1) = 133  ! specific humidity
field_list(field_rttov_q,   2) = 157  ! relative humidity

if ( ozone_data ) then
  field_list(field_rttov_ozone, 1) = 203  ! ozone
endif

if ( clw_data .or. run_scatt .or. ir_addclouds ) then
  field_list(field_rttov_qcl, 1) = 246  ! qcl
end if

! Profiles for running microwave scattering code

if ( run_scatt .or. ir_addclouds ) then
  field_list(field_rttov_qcf, 1)   = 247 ! qcf (12)
  field_list(field_rttov_cfrac, 1) = 248 ! cloud cover (266)
end if

if ( run_scatt ) then
  field_list(field_rttov_rain, 1) = 75 ! Rain mixing ratio
  field_list(field_rttov_snow, 1) = 76 ! Snow mixing ratio
  mmr_snowrain = .true.                ! Specify mass mixing ratio
end if

where (field_list(:,1) > 0)
  got_field = .false.
elsewhere
  got_field = .true.
end where

!-------------------------------------------------------------------------------
! 2. Read GRIB
!-------------------------------------------------------------------------------

! Data is stored as a series of messages, with one field per message. In the
! case of multi-level fields, each level is in a separate message.

print '(a)', 'Reading GRIB messages from ' // trim(file_name)

call grib_open_file(file_id, file_name, 'r')

nmessage = 0

do

  !----------------
  ! Get new message
  !----------------

  call grib_new_from_file(file_id, message_id, status)

  ! Exit if no message found

  if ( status == GRIB_END_OF_FILE ) then
    print '(a)', 'Finished reading: EOF'
    exit
  else if ( status /= GRIB_SUCCESS ) then
    message = 'Could not read GRIB message'
    call radsim_error_report(message, status_error)
  end if

  nmessage = nmessage + 1

  !---------------
  ! Identify field
  !---------------

  call grib_get(message_id, 'centre', centre)
  call grib_get(message_id, 'paramId', paramId)
  call grib_get(message_id, 'dataDate', dataDate)
  call grib_get(message_id, 'dataTime', dataTime)

  if ( output_mode >= output_verbose ) then
    print '(a)',    '---------------'
    print '(2a)',   'centre = ', centre
    print '(a,i0)', 'paramId = ', paramId
    print '(2a)',   'dataDate = ', dataDate
    print '(2a)',   'dataTime = ', dataTime
  end if

  ! Check to see if we need it

  field_index = 0
  do j = 1, size(field_list, dim = 2)
    do i = 1, size(field_list, dim = 1)
      if ( field_list(i,j) == -1 ) cycle
      if ( paramId == field_list(i,j) ) then
        field_index = i
      end if
    end do
  end do

  if ( field_index == 0 ) then
    if ( output_mode >= output_verbose ) print '(a)', 'Skipping this field'
    cycle
  else
    if ( output_mode >= output_verbose ) then
      print '(2a)', 'Getting data for: ', field_names_rttov(field_index)
    end if
  end if

  !---------
  ! Set time
  !---------

  if ( .not. associated(model % validity_time)) then
    allocate(model % validity_time(1,5))
    read(dataDate,'(i4,2i2)') model % validity_time(1,1:3)
    read(dataTime,'(2i2)') model % validity_time(1,4:5)
  end if

  !----------------------
  ! Number of data values
  !----------------------

  ! NV is the number of vertical coordinate coefficients that are stored in pv.
  ! These can be used to construct pressure-level profiles for the ECMWF model
  ! (see below). There are 2 coeffs for each half-level pressure. The number of
  ! full-level pressures is 1 fewer than half-levels.

  call grib_get(message_id, 'numberOfDataPoints', numberOfDataPoints)
  call grib_get(message_id, 'NV', NV)
  call grib_get(message_id, 'level', level)
  call grib_get(message_id, 'missingValue', missingValue)

  call radsim_grib_paramid_name(paramid, param_name)

  if ( output_mode >= output_verbose ) then
    print '(a,i0)', 'numberOfDataPoints = ', numberOfDataPoints
    print '(a,i0)', 'NV = ', NV
    print '(a,i0)', 'level = ', level
  else if ( output_mode == output_default ) then
    print '(a,i0,a,i0,a)', 'Getting level ', level, ' data with paramId: ', &
      paramId, ' (' // trim(param_name) // ')'
  end if

  nprofs = numberOfDataPoints
  nlevels = NV/2 - 1

  !--------------------------
  ! Set pressure level values
  !--------------------------

  ! In the ECMWF model, half-level pressures are a linear function of surface
  ! pressure, with the coefficient values stored in the pv key for
  ! each level of each multi-level field. We only need to do the calculation
  ! once so do it the first time we hit the 2nd level of any field, providing
  ! we already have surface pressure.

  if ( (centre == 'ecmf' .or. centre == 'ecmwf') .and. &
       level > 1 .and. .not. allocated(pv) .and. associated(model % pstar) ) then

    allocate(pv(NV))
    call grib_get(message_id, 'pv', pv, status)

    if ( output_mode >= output_verbose ) print '(a)', 'Calculating pressure levels'

    allocate(model % p(nprofs,nlevels))
    allocate(model % ph(nprofs,nlevels+1))

    call radsim_calc_ecmwf_plevels( &
      model % pstar,   &
      model % p,       &
      model % ph,      &
      pv(1:nlevels+1), &
      pv(nlevels+2:)   )

    got_field(field_rttov_p) = .true.

    ! Convenient place to set this too (we assume it is the same for all fields)
    model % rmdi = missingValue

  end if

  !----------------
  ! Grid definition
  !----------------

  ! The grid definition is available for every field. In most cases this will be
  ! the same but some fields may be on a staggered grid.

  call grib_get(message_id, 'gridType', gridType)

  if ( gridType > 1 ) then
    write(message,'(a,i0,a)') 'This grid type is not supported: ', gridType, &
    ' (see GRIB-2 code table 3.1 for definitions).'
    call radsim_error_report(message, status_error)
  end if

  call grib_get(message_id, 'Ni', Ni)
  call grib_get(message_id, 'Nj', Nj)
  call grib_get(message_id, 'latitudeOfFirstGridPointInDegrees', latitudeOfFirstGridPointInDegrees)
  call grib_get(message_id, 'longitudeOfFirstGridPointInDegrees', longitudeOfFirstGridPointInDegrees)
  call grib_get(message_id, 'latitudeOfLastGridPointInDegrees', latitudeOfLastGridPointInDegrees)
  call grib_get(message_id, 'longitudeOfLastGridPointInDegrees', longitudeOfLastGridPointInDegrees)
  call grib_get(message_id, 'jDirectionIncrementInDegrees', jDirectionIncrementInDegrees)
  call grib_get(message_id, 'iDirectionIncrementInDegrees', iDirectionIncrementInDegrees)
  call grib_get(message_id, 'iScansNegatively', iScansNegatively)
  call grib_get(message_id, 'jScansPositively', jScansPositively)
  call grib_get(message_id, 'jPointsAreConsecutive', jPointsAreConsecutive)

  if ( output_mode >= output_verbose ) then
    print '(a,i0)', 'gridType = ', gridType
    print '(a,i0)', 'Ni = ', Ni
    print '(a,i0)', 'Nj = ', Nj
    print '(a,f8.2)', 'latitudeOfFirstGridPointInDegrees = ', latitudeOfFirstGridPointInDegrees
    print '(a,f8.2)', 'longitudeOfFirstGridPointInDegrees = ', longitudeOfFirstGridPointInDegrees
    print '(a,f8.2)', 'latitudeOfLastGridPointInDegrees = ', latitudeOfLastGridPointInDegrees
    print '(a,f8.2)', 'longitudeOfLastGridPointInDegrees = ', longitudeOfLastGridPointInDegrees
    print '(a,f8.2)', 'jDirectionIncrementInDegrees = ', jDirectionIncrementInDegrees
    print '(a,f8.2)', 'iDirectionIncrementInDegrees = ', iDirectionIncrementInDegrees
    print '(a,i0)', 'iScansNegatively = ', iScansNegatively
    print '(a,i0)', 'jScansPositively = ', jScansPositively
    print '(a,i0)', 'jPointsAreConsecutive = ', jPointsAreConsecutive
  end if

  if (  .not. associated(model % grid % lat) ) then

    model % grid % type = gridType
    model % grid % lon1 = longitudeOfFirstGridPointInDegrees
    model % grid % lat1 = latitudeOfFirstGridPointInDegrees
    model % grid % dlon = abs(iDirectionIncrementInDegrees)
    model % grid % dlat = abs(jDirectionIncrementInDegrees)
    if ( iScansNegatively == 1 ) model % grid % dlon = -model % grid % dlon
    if ( jScansPositively == 0 ) model % grid % dlat = -model % grid % dlat
    if ( jPointsAreConsecutive == 1 ) model % grid % row_order = .false.
    model % grid % ncols = Ni
    model % grid % nrows = Nj

  else

  ! Check to see if grid is different

    if ( model % grid % type /= gridType .or. &
         model % grid % lon1 /= longitudeOfFirstGridPointInDegrees .or. &
         model % grid % lat1 /= latitudeOfFirstGridPointInDegrees .or. &
         abs(model % grid % dlon) /= abs(iDirectionIncrementInDegrees) .or. &
         abs(model % grid % dlat) /= abs(jDirectionIncrementInDegrees) .or. &
         model % grid % ncols /= Ni .or. &
         model % grid % nrows /= Nj ) then

      print '(a)', 'WARNING: GRID IS DIFFERENT!'

    end if

  end if

  !-------------------------------------
  ! Read field into model data structure
  !-------------------------------------

  ! Point to the correct data space, allocating if necessary

  select case(paramId)
    case(3)     ! theta
      if ( .not. associated(model % theta) ) allocate(model % theta(nprofs,nlevels))
      values => model % theta(:,level)
    case(31)    ! Seaice
      if ( .not. associated(model % seaice) ) allocate(model % seaice(nprofs))
      values => model % seaice
    case(54)   ! p on theta levels
      if ( .not. associated(model % p) ) allocate(model % p(nprofs,nlevels))
      values => model % p(:,level)
    case(75)    ! Rain
      if ( .not. associated(model % rain) ) allocate(model % rain(nprofs,nlevels))
      values => model % rain(:,level)
    case(76)    ! Snow
      if ( .not. associated(model % snow) ) allocate(model % snow(nprofs,nlevels))
      values => model % snow(:,level)
    case(130) ! T
      if ( .not. associated(model % t) ) allocate(model % t(nprofs,nlevels))
      values => model % t(:,level)
    case(133)    ! q
      if ( .not. associated(model % q) ) allocate(model % q(nprofs,nlevels))
      values => model % q(:,level)
    case(134, 152)   ! pstar, ln(pstar)
      if ( .not. associated(model % pstar) ) allocate(model % pstar(nprofs))
      values => model % pstar
    case(165)  ! u at 10m
      if ( .not. associated(model % u10) ) allocate(model % u10(nprofs))
      values => model % u10
    case(166)  ! v at 10m
      if ( .not. associated(model % v10) ) allocate(model % v10(nprofs))
      values => model % v10
    case(167)  ! T2
      if ( .not. associated(model % t2) ) allocate(model % t2(nprofs))
      values => model % t2
    case(168)  ! Td2
      if ( .not. associated(model % td2) ) allocate(model % td2(nprofs))
      values => model % td2
    case(172)    ! Land-sea mask
      if ( .not. associated(model % lsm) ) allocate(model % lsm(nprofs))
      values => model % lsm
    case(203)    ! o3
      if ( .not. associated(model % o3) ) allocate(model % o3(nprofs,nlevels))
      values => model % o3(:,level)
    case(235)    ! Tskin
      if ( .not. associated(model % tskin) ) allocate(model % tskin(nprofs))
      values => model % tskin
    case(246)   ! qcl
      if ( .not. associated(model % clw) ) allocate(model % clw(nprofs,nlevels))
      values => model % clw(:,level)
    case(247)    ! qcf
      if ( .not. associated(model % ciw) ) allocate(model % ciw(nprofs,nlevels))
      values => model % ciw(:,level)
    case(129)    ! Geopotential of surface
      if ( .not. associated(model % zsurf) ) allocate(model % zsurf(nprofs))
      values => model % zsurf
    case(248)   ! Cloud fraction
      if ( .not. associated(model % cfrac) ) allocate(model % cfrac(nprofs,nlevels))
      values => model % cfrac(:,level)
    case default
      print '(a)', 'Warning: Not set up to read this field'
  end select

  call grib_get(message_id, 'values', values)
  got_field(field_index) = .true.

  !-----------------
  ! Unit conversions
  !-----------------

  if ( paramId == 129 ) then
    model % zsurf = model % zsurf / gravity  ! Geopotential -> surface elevation
  else if ( paramId == 152 ) then
    model % pstar = exp(model % pstar)       ! ln(psurf) -> psurf
  end if

  call grib_release(message_id)

end do

call grib_close_file(file_id)
if ( allocated(pv) ) deallocate(pv)

!-------------------------------------------------------------------------------
! 3. Derivations
!-------------------------------------------------------------------------------

if ( associated(model % p) ) then
  model % nprofs = size(model % p, dim=1)
  model % nlevels = size(model % p, dim=2)
end if

! LSM from seaice fraction if not present

if ( .not. associated(model % lsm) .and. associated(model % seaice) ) then
  print '(a)', 'No LSM: Deriving LSM from seaice fraction field'
  allocate(model % lsm(nprofs))
  where(model % seaice == model % rmdi)
    model % lsm = 1.0
  elsewhere
    model % lsm = 0.0
  end where
end if

!-------------------------------------------------------------------------------
! 4. Check fields
!-------------------------------------------------------------------------------

do i = 1, size(got_field)
  if ( .not. got_field(i) ) then
    call radsim_error_report( &
      'No data in GRIB file for required RTTOV field: ' // field_names_rttov(i), &
      status_error)
  end if
end do

end subroutine radsim_read_grib
