!-------------------------------------------------------------------------------
! Description:
!
!   Simulate a batch of observations.
!
! Copyright:
!
!   This software was developed within the context of the EUMETSAT Satellite
!   Application Facility on Numerical Weather Prediction (NWP SAF), under the
!   Cooperation Agreement dated 29 June 2011, between EUMETSAT and the
!   Met Office, UK, by one or more partners within the NWP SAF. The partners
!   in the NWP SAF are the Met Office, ECMWF, KNMI and MeteoFrance.
!
!   Copyright 2017, EUMETSAT, All Rights Reserved.
!
!-------------------------------------------------------------------------------

subroutine radsim_run_batch( &
  model, &
  obs, &
  opts, &
  opts_scatt, &
  coeffs, &
  coeffs_scatt, &
  emis_atlas, &
  nc_file_id, &
  startob)

use iso_fortran_env, only : &
  output_unit

use radsim_mod_cfg

use radsim_mod_constants, only : &
  status_error, &
  status_warning, &
  qcflag_domain, &
  qcflag_rttov, &
  qcflag_emiss, &
  real64, &
  int64, &
  int32, &
  output_debug, &
  output_verbose, &
  output_default

use radsim_mod_process, only : &
  nprofs_per_call, &
  z_direction

use radsim_mod_types, only : &
  model_type, &
  obs_type

use mod_rttov_emis_atlas, only : &
  rttov_emis_atlas_data

use rttov_const, only : &
  errorstatus_success, &
  surftype_land, &
  surftype_sea, &
  surftype_seaice, &
  sensor_id_mw

use rttov_types, only : &
  rttov_options, &
  rttov_options_scatt, &
  rttov_coefs, &
  rttov_scatt_coef, &
  rttov_profile, &
  rttov_profile_cloud, &
  rttov_transmission, &
  rttov_radiance, &
  rttov_chanprof, &
  rttov_emissivity

use parkind1, only : jpim, jprb, jplm

implicit none

! radsim routine interfaces

include 'radsim_error_report.interface'
include 'radsim_init_obs_out.interface'
include 'radsim_init_rttov_data.interface'
include 'radsim_model_to_rttov.interface'
include 'radsim_print_ob.interface'
include 'radsim_write_netcdf_obs_1d.interface'
include 'radsim_write_netcdf_obs_nd.interface'

! RTTOV routine interfaces

!include 'rttov_deallocate_emis_atlas.interface'
include 'rttov_direct.interface'
include 'rttov_errorreport.interface'
include 'rttov_get_emis.interface'
include 'rttov_init_prof.interface'
include 'rttov_init_rad.interface'
include 'rttov_init_transmission.interface'
include 'rttov_k.interface'
include 'rttov_print_info.interface'
include 'rttov_print_opts.interface'
include 'rttov_print_profile.interface'
include 'rttov_scatt.interface'

! RTTOV setup variables
type(model_type),          intent(in) :: model
type(obs_type),            intent(inout) :: obs
type(rttov_options),       intent(in) :: opts         ! Options structure
type(rttov_options_scatt), intent(in) :: opts_scatt   ! Options structure for scattering code
type(rttov_coefs),         intent(in) :: coeffs(:)    ! Coefficients structure
type(rttov_scatt_coef),    intent(in) :: coeffs_scatt ! Coefficients structure for scattering code
type(rttov_emis_atlas_data), intent(in) :: emis_atlas
integer(int32),            intent(in) :: nc_file_id
integer, optional,         intent(in) :: startob

! RTTOV variables
integer(kind=jpim) :: nchans
integer(kind=jpim) :: nlevels, nprofs
integer(kind=jpim) :: errorstatus
type(rttov_chanprof),      allocatable :: chanprof(:)
type(rttov_profile),       allocatable :: profiles(:)
type(rttov_profile),       allocatable :: profiles_k(:)
type(rttov_profile_cloud), allocatable :: cld_profiles(:)
type(rttov_emissivity),    allocatable :: emissivity(:)
type(rttov_emissivity),    allocatable :: emissivity_k(:)
logical(kind=jplm),        allocatable :: calcemis(:)
integer(kind=jpim),        allocatable :: frequencies(:)
real(kind=jprb),           allocatable :: emiss(:)
type(rttov_transmission) :: transmission
type(rttov_transmission) :: transmission_k
type(rttov_radiance)     :: radiance
type(rttov_radiance)     :: radiance_k

integer :: i, j, k
integer :: p1, p2, np, prof
integer :: c1, c2
integer :: ncalls
integer :: nprofchans
real    :: tenpercent
integer :: percent
integer :: prof_list(nprofs_per_call)
integer :: batchset
integer(int64) :: array_size
integer(int64) :: array_subdims
integer(int64) :: array_size_max
integer :: nprofs_per_batchset
integer :: level1
integer :: level2
integer :: levelinc
character(len=80) :: message
integer :: start
integer :: obrange(2)
integer :: r1, r2
integer :: callno = 0
integer :: outmode

!-------------------------------------------------------------------------------

callno = callno + 1

!-------------------------------------------------------------------------------
! 1. Set up RTTOV data structures
!-------------------------------------------------------------------------------

nlevels = model % nlevels
nchans = size(obs % channels)

call radsim_init_rttov_data( &
  1_jpim,         & ! in
  nchans,         & ! in
  nlevels,        & ! in
  opts,           & ! in
  opts_scatt,     & ! in
  coeffs,         & ! in
  coeffs_scatt,   & ! in
  chanprof,       & ! inout
  radiance,       & ! inout
  radiance_k,     & ! inout
  transmission,   & ! inout
  transmission_k, & ! inout
  emissivity,     & ! inout
  emissivity_k,   & ! inout
  profiles,       & ! inout
  profiles_k,     & ! inout
  cld_profiles,   & ! inout
  frequencies     ) ! inout

allocate(calcemis(size(emissivity)))
calcemis = .true.

!-------------------------------------------------------------------------------
! 2. Prepare output storage
!-------------------------------------------------------------------------------

! Work out the number of calls required (more than 1 profile may be processed
! in each call).

nprofs = size(model % p, dim=1)

ncalls = nprofs / nprofs_per_call
if ( ncalls*nprofs_per_call /= nprofs ) ncalls = ncalls + 1

! Set instrument details

nprofchans = size(chanprof)
nchans = nprofchans / nprofs_per_call

if ( use_emiss_atlas ) then
  allocate(emiss(nprofchans))
end if

! QC arrays and brightness temperatures or radiances are always required but
! the other fields are optional. If transmittances or Jacobians are requested
! then the output comes from the K code.

if ( .not. associated(obs % qcflags) ) then
  allocate(obs % qcflags(nprofs))
  obs % qcflags = 0
end if

if ( .not. associated(obs % qcrttov) ) then
  allocate(obs % qcrttov(nprofs))
  obs % qcrttov = 0
end if

if ( .not. associated(obs % qcinfo) ) then
  allocate(obs % qcinfo(nprofs))
  obs % qcinfo = 0
end if

allocate(obs % rtsurf(nprofs))
obs % rtsurf = -1

! We need to limit the size of the channel-dependent output arrays to avoid
! running out of virtual memory (size is configurable). This means that
! multi-dimensional arrays in the obs structure have to be written out
! in further batches (batchsets).

array_subdims = nchans
if ( write_trans .or. write_tjac .or. write_qjac .or. write_o3jac ) then
  array_subdims = array_subdims*nlevels
end if

array_size_max = max_array_size*1000*1000
array_size = min(array_size_max, array_subdims*nprofs*4)
nprofs_per_batchset = array_size / (array_subdims*4)

if ( output_mode >= output_default ) then
  print '(a,i0,a)', 'Max output array size = ', array_size, ' bytes'
  print '(a,i0)', 'Number of profs per batchset = ', nprofs_per_batchset
end if

if ( write_radiances ) then
  allocate(obs % radiance(nprofs_per_batchset,nchans))
else
  allocate(obs % bt(nprofs_per_batchset,nchans))
end if

if ( write_emiss ) then
  allocate(obs % emiss(nprofs_per_batchset,nchans))
end if

if ( write_trans ) then
  allocate(obs % trans(nprofs_per_batchset,nlevels,nchans))
end if
if ( write_tjac ) then
  allocate(obs % tjac(nprofs_per_batchset,nlevels,nchans))
end if
if ( write_qjac ) then
  allocate(obs % qjac(nprofs_per_batchset,nlevels,nchans))
end if
if ( write_o3jac ) then
  allocate(obs % o3jac(nprofs_per_batchset,nlevels,nchans))
end if

! Set level-order parameters for Jacobian output

if ( z_direction > 0.0 ) then
  level1 = nlevels
  level2 = 1
  levelinc = -1
else
  level1 = 1
  level2 = nlevels
  levelinc = 1
end if

! Initialise multi-dimensional arrays in the obs structure.

call radsim_init_obs_out(obs)

!-----------------------------------------------------------------------------
! 3. Run simulations
!-----------------------------------------------------------------------------

if ( output_mode >= output_verbose ) then
  print '(a,i0,a)', 'Running rttov with ', nprofs_per_call, ' profiles per call'
end if

p2 = 0
np = 1
prof_list = 0
batchset = 1
if ( present(startob) ) then
  start = startob
else
  start = 1
end if

do i = 1, ncalls

  ! Set up iteration variables. We need the start (p1) and end (p2) positions
  ! within the model field arrays, also the end channel position (c2) within
  ! the RTTOV arrays (=number of channels*number of profiles).

  p1 = p2 + 1
  np = 0
  prof_list = 0

  do j = p1, nprofs
    if ( obs % qcflags(j) == 0 ) then
      np = np + 1
      prof_list(np) = j
      if ( np == nprofs_per_call ) exit
    end if
  end do

  if ( np == 0 ) exit

  p2 = prof_list(np)

  c2 = nchans*np

  !------------------------------------------
  ! 3.1 Transfer model data to RTTOV profiles
  !------------------------------------------

  call radsim_model_to_rttov( &
    prof_list(1:np), & ! in
    model,           & ! in
    obs,             & ! inout
    profiles(1:np),  & ! inout
    cld_profiles     ) ! inout

  if ( output_mode >= output_debug ) then
    do j = 1, np
      print '(a,i0)', 'Profile number ', p1+j-1
      call rttov_print_profile(profiles(j), lu=output_unit)
      call rttov_print_opts(opts, lu=output_unit)
      call rttov_print_info(coeffs(1), lu=output_unit)
    end do
  end if

  ! Assign RT surface type to output structure
  obs % rtsurf(prof_list(1:np)) = profiles(1:np) % skin % surftype

  !-----------------------------
  ! 3.2 Set emissivity variables
  !-----------------------------

  ! Calculate emissivity by default (this should always be the case over sea).

  calcemis = .true.
  emissivity(:) % emis_in = 0.0

  ! Set fixed values for MW over land/seaice if no atlas being used (or if the
  ! atlas returns a bad value). Note these are not likely to be accurate.

  if ( coeffs(1) % coef % id_sensor == sensor_id_mw ) then

    do j = 1, np
      if ( profiles(j) % skin % surftype == surftype_land ) then
        emissivity((j-1)*nchans+1:j*nchans) % emis_in = 0.95
        calcemis((j-1)*nchans+1:j*nchans) = .false.
      else if ( profiles(j) % skin % surftype == surftype_seaice ) then
        emissivity((j-1)*nchans+1:j*nchans) % emis_in = 0.92
        calcemis((j-1)*nchans+1:j*nchans) = .false.
      end if
    end do

  end if

  ! Set input values for land/seaice.

  if ( use_emiss_atlas ) then

    ! Atlas values. It's easier just to retrieve for all profiles but only
    ! assign for land/seaice.

    call rttov_get_emis(  &
      errorstatus,    & ! out
      opts,           & ! in
      chanprof(1:c2), & ! in
      profiles(1:np), & ! in
      coeffs(1),      & ! in
      emis_atlas,     & ! in
      emiss(1:c2)     ) ! out

    c2 = 0

    do j = 1, np

      ! Do not assign for sea

      if ( profiles(j) % skin % surftype == surftype_sea ) then
        c2 = c2 + nchans
        cycle
      end if

      ! Flag if any are bad, otherwise assign to the RTTOV input arrays

      if ( any(emiss((j-1)*nchans+1:j*nchans) <= 0.0) .or. &
           any(emiss((j-1)*nchans+1:j*nchans) >  1.0)      ) then

        obs % qcflags(prof_list(j)) = &
          ibset(obs % qcflags(prof_list(j)), qcflag_emiss)

      else

        emissivity(c2+1:c2+nchans) % emis_in = &
          emiss((j-1)*nchans+1:j*nchans)
        calcemis(c2+1:c2+nchans) = .false.

      end if

      c2 = c2 + nchans

    end do

  end if


  ! Ensure K variables are initialised before each call

  if ( run_k ) then

    call rttov_init_prof(profiles_k)

    emissivity_k % emis_in = 0.
    emissivity_k % emis_out = 0.

    call rttov_init_transmission(transmission_k)

    call rttov_init_rad(radiance_k)
    radiance_k % bt = 1.0

  end if


  !-----------------
  ! 3.3 Run the code
  !-----------------

  if ( np == 0 ) cycle

  if ( .not. run_scatt ) then

    !--------------------
    ! 3.3.1 Standard code
    !--------------------

    if ( .not. run_k ) then

      ! Direct code

      call rttov_direct( &
        errorstatus,          & ! out   error flag
        chanprof(1:c2),       & ! in    channel and profile index structure
        opts,                 & ! in    options structure
        profiles(1:np),       & ! in    profile array
        coeffs(1),            & ! in    coefficients strucutre
        transmission,         & ! inout computed transmittances
        radiance,             & ! inout
        calcemis=calcemis(1:c2),    & ! in
        emissivity=emissivity(1:c2) ) ! inout emissivities

    else

      ! K code

      call rttov_k( &
        errorstatus,      & ! out   error flag
        chanprof(1:c2),   & ! in    channel and profile index structure
        opts,             & ! in    options structure
        profiles(1:np),   & ! in    profile array
        profiles_k(1:c2), & ! inout
        coeffs(1),        & ! in    coefficients strucutre
        transmission,     & ! inout computed transmittances
        transmission_k,   & ! inout
        radiance,         & ! inout
        radiance_k,       & ! inout
        calcemis=calcemis(1:c2),    & ! in
        emissivity=emissivity(1:c2), &
        emissivity_k=emissivity_k(1:c2) ) ! inout emissivities

    end if

  else

    !--------------------------------
    ! 3.3.2 Microwave Scattering Code
    !--------------------------------

    call rttov_scatt ( &
      errorstatus,        & ! out
      opts_scatt,         & ! in
      nlevels,            & ! in
      chanprof(1:c2),     & ! in
      frequencies(1:c2),  & ! in
      profiles(1:np),     & ! in
      cld_profiles(1:np), & ! in
      coeffs(1),          & ! in
      coeffs_scatt,       & ! in
      calcemis(1:c2),     & ! in
      emissivity(1:c2),   & ! inout
      radiance            ) ! out

  end if

  !---------------------------------------------------
  ! 3.4 Transfer outputs to observation data structure
  !---------------------------------------------------

  ! Check error status

  if ( errorstatus /= errorstatus_success ) then
    write(message,'(a,i0)') 'Error in rttov = ', errorstatus
    call radsim_error_report(message, status_warning)
    obs % qcflags(prof_list(1:np)) = &
      ibset(obs % qcflags(prof_list(1:np)), qcflag_rttov)
    obs % qcrttov(prof_list(1:np)) = errorstatus
  end if

  ! Fill output arrays. These are batched by profile so we have to offset the
  ! profile number in assignment.

  do j = 1, np

    prof = prof_list(j) - (batchset-1)*nprofs_per_batchset
    if ( prof > nprofs_per_batchset ) then
      if ( output_mode >= output_verbose ) then
        print '(a,i0)', 'Writing obs output intermediate batch no. ', batchset
      end if
      r1 = 1+(batchset-1)*nprofs_per_batchset
      r2 = batchset*nprofs_per_batchset
      obrange = (/r1, r2/) + start - 1
      call radsim_write_netcdf_obs_nd(nc_file_id, obs, range=obrange)
      call radsim_init_obs_out(obs)
      prof = prof - nprofs_per_batchset
      batchset = batchset + 1
    end if

    c1 = (j-1)*nchans+1
    c2 = j*nchans

    if ( write_radiances ) then
      obs % radiance(prof,:) = radiance % total(c1:c2)
    else
      obs % bt(prof,:) = radiance % bt(c1:c2)
    end if

    if ( write_emiss ) then
      obs % emiss(prof,:) = emissivity(c1:c2) % emis_out
    end if

    if ( associated(obs % trans) ) then
      obs % trans(prof,:,:) = &
        transmission % tau_levels(level1:level2:levelinc, c1:c2)
    end if
    if ( associated(obs % tjac) ) then
      do k = 1, nchans
        obs % tjac(prof,:,k) = profiles_k(k+c1-1) % t(level1:level2:levelinc)
      end do
    end if
    if ( associated(obs % qjac) ) then
      do k = 1, nchans
        obs % qjac(prof,:,k) = profiles_k(k+c1-1) % q(level1:level2:levelinc)
      end do
    end if
    if ( associated(obs % o3jac) ) then
      do k = 1, nchans
        obs % o3jac(prof,:,k) = profiles_k(k+c1-1) % o3(level1:level2:levelinc)
      end do
    end if

    if ( output_mode >= output_debug ) then
      call radsim_print_ob(prof,obs)
    end if

  end do

end do

!-------------------------------------
! 8.5 Write the last batch of obs data
!-------------------------------------

r1 = 1+(batchset-1)*nprofs_per_batchset
r2 = nprofs
obrange = (/r1, r2/) + start - 1

call radsim_write_netcdf_obs_nd(nc_file_id, obs, range=obrange)
call radsim_write_netcdf_obs_1d(nc_file_id, obs, range=(/start,start+nprofs-1/))

! Tidy up

deallocate(calcemis)

call radsim_init_rttov_data( &
  0_jpim,         & ! in
  nchans,         & ! in
  nlevels,        & ! in
  opts,           & ! in
  opts_scatt,     & ! in
  coeffs,         & ! in
  coeffs_scatt,   & ! in
  chanprof,       & ! inout
  radiance,       & ! inout
  radiance_k,     & ! inout
  transmission,   & ! inout
  transmission_k, & ! inout
  emissivity,     & ! inout
  emissivity_k,   & ! inout
  profiles,       & ! inout
  profiles_k,     & ! inout
  cld_profiles,   & ! inout
  frequencies     ) ! inout

end subroutine radsim_run_batch
