!-------------------------------------------------------------------------------
! Description:
!
!   Read RTTOV coefficients and allocate data structures.
!
! Copyright:
!
!   This software was developed within the context of the EUMETSAT Satellite
!   Application Facility on Numerical Weather Prediction (NWP SAF), under the
!   Cooperation Agreement dated 29 June 2011, between EUMETSAT and the
!   Met Office, UK, by one or more partners within the NWP SAF. The partners
!   in the NWP SAF are the Met Office, ECMWF, KNMI and MeteoFrance.
!
!   Copyright 2017, EUMETSAT, All Rights Reserved.
!
!-------------------------------------------------------------------------------

subroutine radsim_setup_rttov( &
  nlevels,     & ! in
  date,        & ! in
  opts,        & ! out
  opts_scatt,  & ! out
  coeffs,      & ! out
  scat_coeffs, & ! out
  emis_atlas   ) ! out

use mod_rttov_emis_atlas, only : &
  rttov_emis_atlas_data

use parkind1, only : &
  jpim, jplm

use radsim_mod_cfg

use radsim_mod_constants, only : &
  status_error, &
  output_verbose

use radsim_mod_process, only : &
  nprofs_per_call

use mod_rttov_emis_atlas, only : &
  atlas_type_mw, &
  atlas_type_ir

use rttov_const, only : &
  errorstatus_success, &
  sensor_id_ir, &
  sensor_id_hi

use rttov_types, only : &
  rttov_options,     &
  rttov_options_scatt, &
  rttov_coefs,       &
  rttov_scatt_coef

implicit none

include 'radsim_error_report.interface'
#include "rttov_errorreport.interface"
#include "rttov_read_coefs.interface"
#include "rttov_read_scattcoeffs.interface"
#include "rttov_setup_emis_atlas.interface"
#include "rttov_user_options_checkinput.interface"

integer(kind=jpim),              intent(in)  :: nlevels
integer(kind=jpim),              intent(in)  :: date(3)
type(rttov_options),             intent(out) :: opts
type(rttov_options_scatt),       intent(out) :: opts_scatt
type(rttov_coefs), allocatable,  intent(out) :: coeffs(:)
type(rttov_scatt_coef),          intent(out) :: scat_coeffs
type(rttov_emis_atlas_data),     intent(out) :: emis_atlas

integer, parameter :: nrttovid = 1        ! Number of sensor coeff files to read

character(len=400) :: file_name_clear
character(len=400) :: file_name_cloud
character(len=80) :: message
integer :: i, j, chan
integer(kind=jpim) :: nchans
integer(kind=jpim) :: nprofchans
integer(kind=jpim) :: errorstatus
integer(kind=jpim) :: file_unit
integer(kind=jpim) :: atlas_type
logical(kind=jplm), allocatable :: usechans(:,:)
integer(kind=jpim) :: year, month

!-------------------------------------------------------------------------------

print '(a)', 'Setting up RTTOV'
print '(a,i0)', ' Number of profiles per call = ', nprofs_per_call

!-------------------------------------------------------------------------------
! 1. Set default options
!-------------------------------------------------------------------------------

! General

opts % config % apply_reg_limits = .true.
opts_scatt % config % apply_reg_limits = .true.
opts % config % verbose = (output_mode >= 2)
opts_scatt % config % verbose = (output_mode >= 2)
opts % rt_all % addrefrac = .false.
opts % rt_all % switchrad = .true.

! MW processing options

opts % rt_mw % clw_data   = clw_data
opts % rt_mw % fastem_version = fastem_version
opts_scatt % fastem_version = fastem_version

! IR processing options

opts % rt_ir % ozone_data = ozone_data
opts % rt_ir % co2_data   = co2_data
opts % rt_ir % n2o_data   = n2o_data
opts % rt_ir % co_data    = co_data
opts % rt_ir % ch4_data   = ch4_data
opts % rt_ir % addsolar   = .false.
opts % rt_ir % addaerosl  = .false.
opts % rt_ir % addclouds  = ir_addclouds
opts % rt_ir % ir_sea_emis_model = ir_sea_emis_model

! Interpolation

opts % interpolation % addinterp = .true.

!-------------------------------------------------------------------------------
! 2. Read coefficients
!-------------------------------------------------------------------------------

errorstatus = 0

!-------------------
! 2.1 Standard model (we always need these)
!-------------------

print '(a)', ' Reading RTTOV coefficient files:'

allocate( coeffs(nrttovid) )

write(unit=file_name_clear, fmt='(a,i0,a)') &
  trim(rttov_coeffs_dir) // '/rtcoef_' // &
  trim(platform) // '_', satid, '_' // trim(inst) // rttov_coeffs_type

print '(a)', '  file = ' // trim(file_name_clear)

if ( ir_addclouds ) then
  write(unit=file_name_cloud, fmt='(a,i0,a)') &
    trim(rttov_sccld_dir) // '/sccldcoef_' // &
    trim(platform) // '_', satid, '_' // trim(inst) // rttov_coeffs_type
  print '(a)', '  file = ' // trim(file_name_cloud)
else
  file_name_cloud = ''
end if

call rttov_read_coefs( &
  errorstatus,                 & ! out
  coeffs(1),                   & ! out
  opts,                        & ! in
  file_coef = file_name_clear, & ! in
  file_sccld = file_name_cloud ) ! in

if ( errorstatus /= errorstatus_success ) then
  write(message, '(a,i0)') 'Error in rttov_read_coefs = ', errorstatus
  call radsim_error_report(message, status_error)
end if


!---------------
! 2.2 Scat model (if requested)
!---------------

if ( run_scatt ) then

  print '(a)', ' Reading coefficient files for RTTOV scat model'

  write(file_name_cloud, fmt='(a)') trim(rttov_mietable_dir) // '/mietable_' // &
    trim(platform) // '_' // trim(inst) // rttov_coeffs_type

  print '(a)', '  file = ' // trim(file_name_cloud)

  file_unit = 10
  open(unit=file_unit, file=file_name_cloud, form='formatted')

  call rttov_read_scattcoeffs( &
    errorstatus,      & ! out
    opts_scatt,       & ! in
    coeffs(1),        & ! in;    clear-sky (direct model) coeffs
    scat_coeffs,      & ! inout; scattering coeffs
    file_id=file_unit ) ! in,    optional; file unit

  close(file_unit)

  if ( errorstatus /= errorstatus_success ) then
    write(message, '(a,i0)') 'Error in rttov_readscattcoeffs = ', errorstatus
    call radsim_error_report(message, status_error)
  end if

end if

!--------------------------------------------------
! 2.3 Check consistency of options and RTTOV coeffs
!--------------------------------------------------

if ( output_mode >= output_verbose ) then
  print '(a)', ' Validating coeffs for requested options'
end if
call rttov_user_options_checkinput(errorstatus, opts, coeffs(1))

if ( errorstatus /= errorstatus_success ) then
  write(message, '(a,i0)') 'Error in rttov_user_options_checkinput = ', errorstatus
  call radsim_error_report(message, status_error)
end if

!---------------------
! 2.4 Emissivity atlas
!---------------------

if ( use_emiss_atlas ) then

  year = date(1)
  month = date(2)

  select case(coeffs(1) % coef % id_sensor)
    case(sensor_id_ir, sensor_id_hi)
      atlas_type = 2
    case default
      atlas_type = 1
  end select

  if ( emiss_atlas_dir /= '' ) then
    call rttov_setup_emis_atlas( &
      errorstatus,     & ! out
      opts,            & ! in
      month,           & ! in
      atlas_type,      & ! in
      emis_atlas,      & ! inout
      emis_atlas_id,   & ! in, optional
      coefs=coeffs(1), & ! in
      path=emiss_atlas_dir ) ! in, optional
  else
    call rttov_setup_emis_atlas( &
      errorstatus,    & ! out
      opts,           & ! in
      month,          & ! in
      atlas_type,     & ! in
      emis_atlas,     & ! inout
      emis_atlas_id,  & ! in, optional
      coefs=coeffs(1) ) ! in
  end if

  if ( errorstatus /= errorstatus_success ) then
    write(message, '(a)') 'Error reading emissivity atlas'
    call radsim_error_report(message, status_error)
  end if

end if

end subroutine radsim_setup_rttov
